#
#include "../param.h"
#include "../conf.h"
#include "../user.h"
#include "../buf.h"

/* control characters */

#define syn	062
#define dle	020
#define ack0	0160
#define ack1	0141
#define enq	055
#define stx	02
#define etb	046
#define etx	03
#define soh	01
#define nak	075
#define wack	0153
#define rvi	0174
#define eot	067
#define pad	0377

/* states of the bsc driver */

#define ttd	011;
#define end 	044
#define bid	01
#define receive	02
#define transmit 03

/* states of the dp registers */

#define trrsta 	03	/* ssync + hduplx */
#define trtsta	0101	/* dtrdy + ienable */
#define rrsta	0103	/* ssync + hduplx + ienable */
#define rtsta	041	/* dtrdy + sienabl */
#define isync	02
#define ring	020000
#define rorun	040000
#define ctrans	0100000
#define done	0200
#define rcvactv	04000

#define dpb	0174760
#define kgadr	0170700
#define bspri	-1

/* timeouts */

#define timebid	 50
#define timewack 80
#define timettd	 80
#define timeansw 150
#define timeactv 1000


/* dp registers */

struct
	{
	int  dprcsr;
	char dprbuf;
	char dpsyn0;
	int  dptcsr;
	char dptbuf;
	char dpsyn1;
	};


/* input/output buffers */

struct besbuf
	{
	char beshead[3];
	char stxhead;
	char besbf[516];
	int  bescount;
	char bessta;
	char dpextra[77];	/* total 600*2 for use with dp.c */
	}
besbpool[2],*besbp;

/* input/output queues */

struct
	{
	struct besbuf *bsrb1;
	struct besbuf *bsrb2;
	struct besbuf *bsrp1;
	struct besbuf *bsrp2;
	}
bsreq;

/* kg registers */

struct
	{
	char kgstat;
	char kgunused;
	int  kgbccr;
	int  kgdata;
	};


/* control information */
struct
	{
	int  bstime;
	char *bsrpnt;
	int  bsrcnt;
	char *bswpnt;
	int  bswcnt;
	char bssta;
	char bsrsta;
	char bswsta;
	char bsrerr;
	char bswerr;
	char bsror;
	char bsetbx;
	char bsenqc;
	char bsackx;
	char bsdle;
	char bspnt;
	char bssave;
	char bsansw;
	}
bsdr;
struct buf bscbuf;    /*fake buffer header to let bsc use iomove */
char bsr1[]{syn,syn,syn,enq,pad};
char bsr2[]{syn,syn,syn,dle,ack0,pad};

bsopen(dev,flag)
{
if(bsdr.bssta)
		{
		u.u_error=ENXIO;  /* device already opened */
		return;
		}
besbpool[0].beshead[0]=besbpool[0].beshead[1]=besbpool[0].beshead[2]=syn;
besbpool[1].beshead[0]=besbpool[1].beshead[1]=besbpool[1].beshead[2]=syn;
besbpool[0].stxhead=besbpool[1].stxhead=stx;
bsdr.bsrsta=bsdr.bswsta=0;
bsdr.bswerr=bsdr.bsrerr=0;
dpb->dpsyn0=syn;
bsclear(flag);
if(flag)
	{
	/* start bidding */
	dpb->dprcsr=trrsta;
	dpb->dptcsr=trtsta;
	bsr1[3]=enq;
	bsdr.bswpnt=bsr1;
	bsdr.bswcnt=5;
	bsout();
	dpb->dptcsr=| isync;
	}
else
	{
	dpb->dprcsr=rrsta;
	dpb->dptcsr=rtsta;
	}
}

bsclose(dev)
{
spl6();
dpb->dprcsr=dpb->dptcsr= 0;
bsdr.bsrerr=bsdr.bswerr= ENOEXEC;
bsdr.bssta=0;
wakeup(&bsdr);
spl0();
}

bstty(dev,av)
int *av;
{
register *v;
if(v=av)
	{
	*v++ = dpb->dptcsr;
	*v++ =0;
	*v=0;
	}
}


bsread()
{
while(bsdr.bssta != receive)
			{
			sleep(&bsdr,bspri);
			if(u.u_error = bsdr.bsrerr)return;
			}
spl6();
while((besbp= bsreq.bsrp1) == 0)
				{
				if(bsdr.bsrsta == end)
							{
							bsdr.bsrsta=0;
							spl0();
							return;
							}
				sleep(&bsdr,bspri);
				if(u.u_error=bsdr.bsrerr)
							{
							spl0();
							return;
							}
				}
spl0();
if(bsdr.bsrsta == end)
			{
			bsdr.bsrsta=0;
			return;
			}
bscbuf.b_addr= &besbp->besbf[0];
iomove(&bscbuf,0,besbp->bescount,B_READ);
spl6();
bsreq.bsrp1=bsreq.bsrp2;
bsreq.bsrp2=0;
if(bsreq.bsrb1)bsreq.bsrb2=besbp;
else		bsreq.bsrb1=besbp;
spl0();
}

bswrite()
{
register int c;
int bcc;

if((c=u.u_count) > 514)
			{
			u.u_error=ENOSPC;
			spl0();
			return;
			}
spl6();
if(u.u_error=bsdr.bswerr)return;
while((besbp=besgetb()) == 0)
				{
				sleep(&bsdr,bspri);
				if(u.u_error=bsdr.bswerr)
							{
							spl0();
							return;
							}
				}
spl0();
bscbuf.b_addr= &besbp->besbf[0];
iomove(&bscbuf,0,c,B_WRITE);
bcc=kgsubr(&besbp->besbf[0],c);
besbp->besbf[c++]=bcc.lobyte;
besbp->besbf[c++]=bcc.hibyte;
besbp->besbf[c]=pad;
besbp->bescount=c+5;
spl6();
if(u.u_error=bsdr.bswerr)
			{
			spl0();
			return;
			}
if(bsreq.bsrb1)bsreq.bsrb2=besbp;
else	       bsreq.bsrb1=besbp;
if(besbp->besbf[c-3] == etx)
				{
				sleep(&bsdr,bspri);
				u.u_error=bsdr.bswerr;
				}
spl0();
}

struct besbuf *besgetb()
{
register int i;
for(i=0;i < 2;i++)
if(besbpool[i].bessta == 0)
			{
			besbpool[i].bessta=1;
			return(&besbpool[i]);
			}
return(0);
}

bsrint()
{
bsdr.bssave=dpb->dprbuf;
if(++bsdr.bsrcnt == 1)
			{
			if(bsdr.bssave == dle)
						{
						bsdr.bsdle=1;
						return;
						}
			if(bsdr.bssta == bid)
			/* format error */
						{
						bsferr();
						return;
						}
		if(bsdr.bssta == transmit)
				{
				if(bsdr.bssave == nak)
				/* nak received */
						{
						bstran();
						return;
						}
		if(((bsdr.bssave == stx)||(bsdr.bssave == soh))&&(bsdr.bswsta == etx))
			/* conversational acknowledjement */
						{
						bschange();
						return;
						}
				if(bsdr.bssave == eot)
				/* eot received */
							{
							bseotr(0);
							return;
							}
				bsferr();
				return;
				}
		if((bsdr.bssave == stx)||(bsdr.bssave == soh))return;
						/* stx or soh received */
		if(bsdr.bssave == enq)
		/* enq received */
					{
					if(bsdr.bsansw == nak)bsresp();
					else		      bswack();
					return;
					}
		if(bsdr.bssave == eot)
		/* eot received */
					{
					if(bsdr.bsrsta == etx)
							{
							bsdr.bsrsta=end;
							bsclear(0);
							wakeup(&bsdr);
							return;
							}
					else
					   {
					   bseotr(0);
					   return;
					   }
					}
		/* format error */
		bsfrerr();
		return;
			}
if(bsdr.bsrcnt == 2)
		{
		if(bsdr.bssta == bid)
				{
				if(bsdr.bssave == ack0)
				/* ack0 received after bidding */
							{
							bsdr.bssta=transmit;
							bsdr.bsackx=ack1;
							bstran();
							}
				/* format error */
				else bsferr();
				return;
				}
		if(bsdr.bsdle && (bsdr.bssave == eot))
			/* dle eot received */
							{
							bseotr(1);
							return;
							}
		if(bsdr.bssta == transmit)
		/* dle xxx received */
					{
					bstest();
					return;
					}
		}
if(bsdr.bsrcnt > 516)
/* too long block */
		{
		bsfrerr();
		return;
		}
*bsdr.bsrpnt++ = bsdr.bssave;
if(bsdr.bsetbx == 1)
/* first crc byte */
		{
		bsdr.bsetbx=2;
		return;
		}
if(bsdr.bsetbx == 2)
/* second crc byte */
		{
		bsdr.bsetbx=0;
		testfr();
		return;
		}
if((bsdr.bssave == etb)||(bsdr.bssave == etx))
	/* etb or etx received */
					{
					bsdr.bsetbx=1;
					return;
					}
if(bsdr.bssave == enq)
/* stx ---- enq received */
		{
		bsdr.bsansw=nak;
		bsresp();
		return;
		}
}

bswint()
{
register int dpstat;
dpstat=dpb->dptcsr;
dpb->dptcsr=& ~(rorun | ring | ctrans | done);
if(dpstat & ring)return;
if(dpstat & rorun)
		{
		bsdr.bsror++;
		if(bsdr.bsror == 5)
				{
				bsdr.bsror=0;
				printf("5 roruns on bs \n");
				}
		bsfrerr();
		return;
		}
if(dpstat & ctrans)
		{
		bsfrerr();
		return;
		}

bsout();
}

bsout()
{
int bstout();
register char c;
if(--bsdr.bswcnt >= 0)
			{
			c= *bsdr.bswpnt++;
			c=& 0377;
			dpb->dptbuf=c;
			if(bsdr.bswcnt == 0)dpb->dprcsr=& ~isync;
			return;
			}
++bsdr.bspnt;
if(bsdr.bssta == transmit)
			{
			if((bsdr.bswerr) || (bsdr.bswsta == end))
							{
							bsclear(0);
							wakeup(&bsdr);
							return;
							}
			bsdr.bstime=timeansw;
			}
else
if(bsdr.bssta == bid)bsdr.bstime=timebid;
else
if(bsdr.bssta == receive)bsdr.bstime=timeactv;
dpb->dprcsr=rrsta;
dpb->dptcsr=rtsta;
timeout(bstout,bsdr.bspnt,bsdr.bstime);
}


kgsubr(kgstring,kglng)
int *kgstring;
{
register int *stringp;
stringp= kgstring;
kgadr->kgstat=0131;
for(;kglng > 1;kglng=- 2)kgadr->kgdata= *stringp++;
if(kglng)
	{
	kgadr->kgstat=0101;
	kgadr->kgdata= *stringp;
	}
return(kgadr->kgbccr);
}

bstout(point)
{
if(bsdr.bssta == 0)return;
if(point != bsdr.bspnt)return;
if(bsdr.bssta == bid)
 /* no answer to the bidding */
			{
			bsdr.bswpnt=bsr1;
			bsdr.bswcnt=5;
			bsout();
			dpb->dptcsr=| isync;
			return;
			}
if(bsdr.bssta == transmit)
			{
			if(bsdr.bstime == timettd)
						{
						if(bsreq.bsrb1)
							      {
       							      bstran();
							      return;
							      }
						/* send ttd */
						bsr2[3]=stx;
						bsr2[4]=enq;
						bsdr.bswpnt=bsr2;
						bsdr.bswcnt=6;
						bsout();
						dpb->dptcsr=| isync;
						return;
						}
			if(bsdr.bstime == timeansw)
			 /* no answer in 3 seconds */
						{
						bsferr();
						return;
						}
			}
if(bsdr.bstime == timeactv)
 /* 20 seconds no activities */
			{
			bsdr.bsrerr=ENOEXEC;
			bsclear(0);
			wakeup(&bsdr);
			return;
			}
if(bsreq.bsrb1)
		{
		bswack();
		return;
		}
/* send wack */
bsr2[3]=dle;
bsr2[4]=wack;
bsdr.bswpnt=bsr2;
bsdr.bswcnt=6;
bsout();
dpb->dptcsr=| isync;
}

bstclear()
{
++bsdr.bspnt;
dpb->dprcsr=trrsta;
dpb->dptcsr=trtsta;
bsdr.bsdle=0;
}

bsferr()
{
bstclear();
if(bsdr.bsenqc++ > 15)
			{
			/* send eot */
			bsr1[3]=eot;
			bsdr.bswerr=EIO;
			bsdr.bswsta=eot;
			}
else
    /* send enq */
    bsr1[3]=enq;
bsdr.bswpnt=bsr1;
bsdr.bswcnt=5;
bsout();
dpb->dptcsr=| isync;
}

bsfrerr()
{
++bsdr.bspnt;
dpb->dprcsr=& ~rcvactv;
bsdr.bsrpnt= &bsreq.bsrb1->besbf[0];
bsdr.bsrcnt=0;
bsdr.bstime=timeactv;
bsdr.bsdle=0;
timeout(bstout,bsdr.bspnt,bsdr.bstime);
}

bseotr(par)
{
++bsdr.bspnt;
if(par)
	{
	dpb->dprcsr=dpb->dptcsr=0;
	bsdr.bswerr=bsdr.bsrerr=ENXIO;
	}
else
	{
	if(bsdr.bssta== receive)bsdr.bsrerr=ENOEXEC;
	else
	     bsdr.bswerr=ENOEXEC;
	bsclear(0);
	}
wakeup(&bsdr);
}

bstran()
{
bstclear();
if(bsreq.bsrb1)
		{
		 /* send a block */
		bsdr.bswpnt= &bsreq.bsrb1->beshead[0];
		bsdr.bswcnt= bsreq.bsrb1->bescount;
		if(bsreq.bsrb1->besbf[bsreq.bsrb1->bescount-8] == etx)
							bsdr.bswsta=etx;
		bsout();
		dpb->dptcsr=| isync;
		return;
		}
else
		if(bsdr.bswsta == rvi)
					{
					/* send eot after rvi received */
					bsdr.bswerr=ETXTBSY;
					bsr1[3]=eot;
					bsdr.bswpnt=bsr1;
					bsdr.bswcnt=5;
					bsout();
					dpb->dptcsr=| isync;
					return;
					}
bsdr.bswsta=ttd;
bsdr.bstime=timettd;
timeout(bstout,bsdr.bspnt,bsdr.bstime);
}

bstest()
{
if((bsdr.bssave == bsdr.bsackx) || (bsdr.bssave == rvi))
		{
		if(bsdr.bswsta == etx)
					{
					/* send eot after etx */
					bsdr.bswsta=end;
					bsdr.bswpnt=bsr1;
					bsr1[3]=eot;
					bsdr.bswcnt=5;
					bsout();
					dpb->dptcsr=| isync;
					return;
					}
		if(bsdr.bssave == rvi)
					{
					bsdr.bswsta=rvi;
					if(bsdr.bsackx == ack0)bsdr.bsackx=ack1;
					else			bsdr.bsackx=ack0;
					bsreq.bsrb1=bsreq.bsrb2;
					bsreq.bsrb2=0;
					bstran();
					return;
					}
		if(bsdr.bsackx == ack0)bsdr.bsackx=ack1;
		else			bsdr.bsackx=ack0;
		bsreq.bssta=0;
		bsreq.bsrb1=bsreq.bsrb2;
		bsreq.bsrb2=0;
		if(bsdr.bswsta != rvi)wakeup(&bsdr);
		bstran();
		return;
		}
if(bsdr.bssave == wack)
			{
			/* send enq */
			bsr1[3]=enq;
			bsdr.bswpnt=bsr1;
			bsdr.bswcnt=5;
			bsout();
			dpb->dptcsr=| isync;
			return;
			}
bsferr();
}

bsresp()
{
if(bsdr.bsansw == nak)
			{
			 /* send nak */
			bsdr.bsrpnt=bsreq.bsrb1->besbf[0];
			bsdr.bsrcnt=0;
			bsdr.bswcnt=5;
			bsr1[3]=nak;
			bsdr.bswpnt=bsr1;
			}
else
			{
			/* send acknowledjement */
			bsr2[3]=dle;
			bsr2[4]=bsdr.bsansw;
			bsdr.bswpnt=bsr2;
			bsdr.bswcnt=6;
			}
bsout();
dpb->dptcsr=| isync;
}

testfr()
{
bstclear();
if(kgsubr(&bsreq.bsrb1->besbf[0],bsdr.bsrcnt-1))
						{
						bsdr.bsansw=nak;
						bsresp();
						}
else

	{
	bsreq.bsrb1->bescount=bsdr.bsrcnt-4;
	if(bsreq.bsrb1->besbf[bsdr.bsrcnt-4] == etx)bsdr.bsrsta=etx;
	else					    bsdr.bswsta=0;
	bsdr.bsansw=bsdr.bsackx;
	if(bsdr.bsackx == ack0)bsdr.bsackx=ack1;
	else		       bsdr.bsackx=ack0;
	if(bsreq.bsrp1)bsreq.bsrp2=bsreq.bsrb1;
	else		bsreq.bsrp1=bsreq.bsrb1;
	bsreq.bsrb1=bsreq.bsrb2;
	bsreq.bsrb2=0;
	wakeup(&bsdr);
	bswack();
	}
}

bswack()
{
if(bsreq.bsrb1)
		{
		bsdr.bsrsta=0;
		bsdr.bsrpnt= &bsreq.bsrb1->besbf[0];
		bsdr.bsrcnt=0;
		bsresp();
		}
else
	{
	bsdr.bstime=timewack;
	bsdr.bsrsta=wack;
	timeout(bstout,bsdr.bspnt,bsdr.bstime);
	}
}

bschange()
{
bsclear(0);
bsdr.bswsta=0;
bsdr.bsackx=ack1;
bsdr.bsrcnt=1;
wakeup(&bsdr);
}

bsclear(flag)
{
bsdr.bsenqc=bsdr.bsror=bsdr.bsetbx=0;
bsdr.bsdle=0;
bsdr.bsackx=ack1;
besbpool[0].bessta=besbpool[1].bessta=0;
if(flag)
	{
	bsdr.bssta= bid;
	bsdr.bsrb1=bsreq.bsrb2=0;
	}
else
	
	{
	bsdr.bssta=receive;
bsdr.bsansw=ack0;
	bsreq.bsrb1= &besbpool[0];
	bsreq.bsrb2= &besbpool[1];
	bsdr.bsrpnt= &besbpool[0].besbf[0];
	bsdr.bsrcnt=0;
	bsreq.bsrp1=bsreq.bsrp2=0;
	bsdr.bstime=timeactv;
	++bsdr.bspnt;
	timeout(bstout,bsdr.bspnt,bsdr.bstime);
	}
}
